#include <comic_spe.h>
#include "jacobi01_common.h"

COMIC_shared_extern();
COMIC_shared_region();

static inline void jacobi01_spe_f1();
static inline void jacobi01_spe_f2();


/***********************************************/
/* SPE main function                           */
/***********************************************/
COMIC_spe_main(
	case 1: jacobi01_spe_f1(); break;
	case 2: jacobi01_spe_f2(); break;
)


static inline void jacobi01_spe_f1()
{
	int tid = COMIC_get_spe_thread_id();

	int i, j;
	int idx;
	int m, m_factor, m_start, m_end, n;

	m = LoadInt( &COMIC_access(m) );
	n = LoadInt( &COMIC_access(n) );

	int f1_iter_start, f1_iter_end;
	unsigned int addr_range, addr_amount;

	addr_range  = (unsigned int)&COMIC_access(error) - (unsigned int)&COMIC_access(uold);
	addr_amount = addr_range / (COMIC_MAXSPE+1);
	if( (addr_amount & ~COMIC_get_page_mask()) < addr_amount )
	 	addr_amount = (addr_amount + COMIC_get_page_size()) & ~COMIC_get_page_mask();
	f1_iter_start = (addr_amount * (tid+1)) / sizeof(double);
	f1_iter_end = f1_iter_start + (addr_amount / sizeof(double)) - 1;

	double u;
	double *p_global;
	double *u_local, *uold_local;
	vec_uint4 v_pb_u = spu_splats( (unsigned int)0 );
	vec_uint4 v_pb_uold = spu_splats( (unsigned int)0 );

	int lcnt = COMIC_get_page_size() / sizeof(double);
	int pcnt = (f1_iter_end - f1_iter_start + 1) / lcnt;

	for( i=0; i<pcnt; i++ ) {
		idx = f1_iter_start + i * lcnt;

		p_global = &COMIC_access(u[idx]);
		v_pb_u = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
						COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
		u_local = RODoublePtr( p_global );

		p_global = &COMIC_access(uold[idx]);
		v_pb_uold = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
						COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
		uold_local = (double*)COMIC_spe_write_single_LSA( (void*)p_global );			

		for( j=0; j<lcnt; j++ ) {
			*uold_local = *u_local;
			u_local++;
			uold_local++;
		}
	}

	COMIC_spe_barrier( 0 );
}


static inline void jacobi01_spe_f2()
{
	int tid = COMIC_get_spe_thread_id();

	short i, j;
  double omega, resid, ax, ay, b;
	double error = 0.0;
	double dbl_temp;
	int idx;
	int m, m_factor, m_start, m_end, n;

	m = LoadInt( &COMIC_access(m) );
	n = LoadInt( &COMIC_access(n) );
	omega = LoadDouble( &COMIC_access(omega) );
	ax = LoadDouble( &COMIC_access(ax) );
	ay = LoadDouble( &COMIC_access(ay) );
	b = LoadDouble( &COMIC_access(b) );
	
	int f2_iter_start, f2_iter_end;
	int f2_j_start, f2_j_end, f2_i_start, f2_i_end;
	unsigned int addr_range, addr_amount;
	addr_range = (unsigned int)&COMIC_access(f) - (unsigned int)&COMIC_access(u);
	addr_amount = addr_range / (COMIC_MAXSPE+1);
	if( (addr_amount & ~COMIC_get_page_mask()) < addr_amount )
	 	addr_amount = (addr_amount + COMIC_get_page_size()) & ~COMIC_get_page_mask();
	f2_iter_start = (addr_amount * (tid+1)) / sizeof(double);
	if( (tid+1) == 0 ) f2_j_start = 1;
	else               f2_j_start = f2_iter_start / n;
	f2_i_start = f2_iter_start % n;

	if( tid == (COMIC_MAXSPE - 1) ) 
		addr_amount = addr_range - COMIC_MAXSPE * addr_amount;
	f2_iter_end = f2_iter_start + (addr_amount / sizeof(double)) - 1;
	if( tid == (COMIC_MAXSPE - 1) ) {
		f2_j_end = m - 2;
		f2_i_end = n - 2;
	}
	else {
		f2_j_end = f2_iter_end / n;
		f2_i_end = f2_iter_end % n;
	}
	m_start = f2_j_start;
	m_end   = f2_j_end;


	double uold, uold_m1, uold_p1, uold_mM, uold_pM, f;
	double *p_global;
	double *uold_m1_local;
	double *uold_mM_local;
	double *uold_pM_local;
	double *f_local, *u_local;
	vec_uint4 v_pb_uold_m1 = spu_splats( (unsigned int)0 );
	vec_uint4 v_pb_uold_mM = spu_splats( (unsigned int)0 );
	vec_uint4 v_pb_uold_pM = spu_splats( (unsigned int)0 );
	vec_uint4 v_pb_f = spu_splats( (unsigned int)0 );
	vec_uint4 v_pb_u = spu_splats( (unsigned int)0 );

	/* preamble */
	if( f2_i_start > 1 ) {
		j = f2_j_start;
		for( i=f2_i_start; i<n-1; i++ ) {
			idx = i+m*j;
			
			if( __builtin_expect( (unsigned int)&COMIC_access(uold[idx+1]) < spu_extract(v_pb_uold_m1,0), 1 ) ) {
				uold_m1_local++;
				uold_m1 = *uold_m1_local;
				uold    = *(uold_m1_local + 1);
				uold_p1 = *(uold_m1_local + 2);
			}
			else {
				p_global = &COMIC_access(uold[idx-1]);
				v_pb_uold_m1 = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								               COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				uold_m1_local = RODoublePtr( p_global );
				uold_m1 = *uold_m1_local;
				uold    = RODouble( &COMIC_access(uold[idx]) );
				uold_p1 = RODouble( &COMIC_access(uold[idx+1]) );
			}

			p_global = &COMIC_access(uold[idx-m]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_uold_mM,0), 1 ) ) {
				uold_mM_local++;
			}
			else {
				v_pb_uold_mM = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								               COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				uold_mM_local = RODoublePtr( p_global );
			}
			uold_mM = *uold_mM_local;

			p_global = &COMIC_access(uold[idx+m]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_uold_pM,0), 1 ) ) {
				uold_pM_local++;
			}
			else {
				v_pb_uold_pM = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								               COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				uold_pM_local = RODoublePtr( p_global );
			}
			uold_pM = *uold_pM_local;

			p_global = &COMIC_access(f[idx]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_f,0), 1 ) ) {
				f_local++;
			}
			else {
				v_pb_f = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								         COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				f_local = RODoublePtr( p_global );
			}
			f = *f_local;


			resid = ( ax * (uold_m1 + uold_p1)
				+ ay * (uold_mM + uold_pM)
				+ b * uold - f
				) / b;
			//resid = ( ax * (RODouble( &COMIC_access(uold[idx-1]) ) + RODouble( &COMIC_access(uold[idx+1]) ))
			//	+ ay * (RODouble( &COMIC_access(uold[idx-m]) ) + RODouble( &COMIC_access(uold[idx+m]) ))
			//	+ b * uold - RODouble( &COMIC_access(f[idx]) )
			//	) / b;

			// update solution 
			p_global = &COMIC_access(u[idx]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_u,0), 1 ) ) {
				u_local++;
			}
			else {
				v_pb_u = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								         COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				u_local = (double*)COMIC_spe_write_single_LSA( (void*)p_global );
			}
			*u_local = uold - omega * resid;
			//SWDouble( &COMIC_access(u[idx]), uold - omega * resid );

			// accumulate residual error
			error = error + resid*resid;
	  }

		m_start = f2_j_start + 1;
	}
	if( f2_i_end < n - 2 ) {
		m_end = f2_j_end - 1;
	}

	/* main */
	//vec_uint4 v_pb = spu_splats( (unsigned int)0 );
	//vec_uint4 v_gptr, v_cmp;
	for (j=m_start; j<=m_end; j++) {
		v_pb_uold_m1 = spu_splats( (unsigned int)0 );
		v_pb_uold_mM = spu_splats( (unsigned int)0 );
		v_pb_uold_pM = spu_splats( (unsigned int)0 );
		v_pb_f = spu_splats( (unsigned int)0 );
		v_pb_u = spu_splats( (unsigned int)0 );
		//v_pb = spu_splats( (unsigned int)0 );

	  for (i=1; i<n-1; i++){
			idx = i+m*j;
			
			if( __builtin_expect( (unsigned int)&COMIC_access(uold[idx+1]) < spu_extract(v_pb_uold_m1,0), 1 ) ) {
				uold_m1_local++;
				uold_m1 = *uold_m1_local;
				uold    = *(uold_m1_local + 1);
				uold_p1 = *(uold_m1_local + 2);
			}
			else {
				p_global = &COMIC_access(uold[idx-1]);
				v_pb_uold_m1 = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								               COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				uold_m1_local = RODoublePtr( p_global );
				uold_m1 = *uold_m1_local;
				uold    = RODouble( &COMIC_access(uold[idx]) );
				uold_p1 = RODouble( &COMIC_access(uold[idx+1]) );
			}

			p_global = &COMIC_access(uold[idx-m]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_uold_mM,0), 1 ) ) {
				uold_mM_local++;
			}
			else {
				v_pb_uold_mM = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								               COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				uold_mM_local = RODoublePtr( p_global );
			}
			uold_mM = *uold_mM_local;

			p_global = &COMIC_access(uold[idx+m]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_uold_pM,0), 1 ) ) {
				uold_pM_local++;
			}
			else {
				v_pb_uold_pM = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								               COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				uold_pM_local = RODoublePtr( p_global );
			}
			uold_pM = *uold_pM_local;

			p_global = &COMIC_access(f[idx]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_f,0), 1 ) ) {
				f_local++;
			}
			else {
				v_pb_f = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								         COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				f_local = RODoublePtr( p_global );
			}
			f = *f_local;

			resid = ( ax * (uold_m1 + uold_p1)
				+ ay * (uold_mM + uold_pM)
				+ b * uold - f
				) / b;
			//resid = ( ax * (RODouble( &COMIC_access(uold[idx-1]) ) + RODouble( &COMIC_access(uold[idx+1]) ))
			//	+ ay * (RODouble( &COMIC_access(uold[idx-m]) ) + RODouble( &COMIC_access(uold[idx+m]) ))
			//	+ b * uold - RODouble( &COMIC_access(f[idx]) )
			//	) / b;

			// update solution 
			p_global = &COMIC_access(u[idx]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_u,0), 1 ) ) {
				u_local++;
			}
			else {
				v_pb_u = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								         COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				u_local = (double*)COMIC_spe_write_single_LSA( (void*)p_global );
			}
			*u_local = uold - omega * resid;
			//SWDouble( &COMIC_access(u[idx]), uold - omega * resid );

			// accumulate residual error
			error = error + resid*resid;
	  }
	}

	/* epilogue */
	if( f2_i_end < n - 2 ) {
		v_pb_uold_m1 = spu_splats( (unsigned int)0 );
		v_pb_uold_mM = spu_splats( (unsigned int)0 );
		v_pb_uold_pM = spu_splats( (unsigned int)0 );
		v_pb_f = spu_splats( (unsigned int)0 );
		v_pb_u = spu_splats( (unsigned int)0 );

		j = f2_j_end;
		for( i=1; i<=f2_i_end; i++ ) {
			idx = i+m*j;
			
			if( __builtin_expect( (unsigned int)&COMIC_access(uold[idx+1]) < spu_extract(v_pb_uold_m1,0), 1 ) ) {
				uold_m1_local++;
				uold_m1 = *uold_m1_local;
				uold    = *(uold_m1_local + 1);
				uold_p1 = *(uold_m1_local + 2);
			}
			else {
				p_global = &COMIC_access(uold[idx-1]);
				v_pb_uold_m1 = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								               COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				uold_m1_local = RODoublePtr( p_global );
				uold_m1 = *uold_m1_local;
				uold    = RODouble( &COMIC_access(uold[idx]) );
				uold_p1 = RODouble( &COMIC_access(uold[idx+1]) );
			}

			p_global = &COMIC_access(uold[idx-m]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_uold_mM,0), 1 ) ) {
				uold_mM_local++;
			}
			else {
				v_pb_uold_mM = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								               COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				uold_mM_local = RODoublePtr( p_global );
			}
			uold_mM = *uold_mM_local;

			p_global = &COMIC_access(uold[idx+m]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_uold_pM,0), 1 ) ) {
				uold_pM_local++;
			}
			else {
				v_pb_uold_pM = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								               COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				uold_pM_local = RODoublePtr( p_global );
			}
			uold_pM = *uold_pM_local;

			p_global = &COMIC_access(f[idx]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_f,0), 1 ) ) {
				f_local++;
			}
			else {
				v_pb_f = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								         COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				f_local = RODoublePtr( p_global );
			}
			f = *f_local;


			resid = ( ax * (uold_m1 + uold_p1)
				+ ay * (uold_mM + uold_pM)
				+ b * uold - f
				) / b;
			//resid = ( ax * (RODouble( &COMIC_access(uold[idx-1]) ) + RODouble( &COMIC_access(uold[idx+1]) ))
			//	+ ay * (RODouble( &COMIC_access(uold[idx-m]) ) + RODouble( &COMIC_access(uold[idx+m]) ))
			//	+ b * uold - RODouble( &COMIC_access(f[idx]) )
			//	) / b;

			// update solution 
			p_global = &COMIC_access(u[idx]);
			if( __builtin_expect( (unsigned int)p_global < spu_extract(v_pb_u,0), 1 ) ) {
				u_local++;
			}
			else {
				v_pb_u = spu_sl( spu_add( spu_rlmask( spu_promote((unsigned int)p_global,0), 
								         COMIC_get_pz_log2_m1() ), 1 ), COMIC_get_pz_log2() );
				u_local = (double*)COMIC_spe_write_single_LSA( (void*)p_global );
			}
			*u_local = uold - omega * resid;
			//SWDouble( &COMIC_access(u[idx]), uold - omega * resid );

			// accumulate residual error
			error = error + resid*resid;
	  }
	}

	SWDouble( &COMIC_access(error[tid*COMIC_get_page_size()]), error );

	COMIC_spe_barrier( 0 );
}
